int DEBUGDV=0; // set to 1 for debug information

/* General function for Deal */
void Deal (int nw, Imode_t imode, eSkel_molecule_t *worker (eSkel_molecule_t *),
	   int col, Dmode_t dmode,
	   void *in, int inlen, int inmul, spread_t inspr, MPI_Datatype inty,
	   void *out, int outlen, int *outmul, spread_t outspr, MPI_Datatype outty,
	   int outbuffsz, MPI_Comm comm)
{

  int i, ROOT;
  int *givelen;
  eSkel_molecule_t *tempitemin;
  eSkel_molecule_t *tempitemout;
  int **len;
  int *outlenptr;
  MPI_Group stagerootonlygroup, tempgroup;

  if (DEBUGDV) printf("Entering Deal, rank %d, col %d, Top %d --", myrank(), col, Top+1);

  Top++;

  // Some of this copying is probably redundant
  FrStack[Top]             = (skel_t *) malloc (sizeof(skel_t));
  FrStack[Top]->skel       = Dskel;
  FrStack[Top]->dmode      = dmode;   
  FrStack[Top]->mycol      = col;   
  FrStack[Top]->mycalltree = NULL;    // for now. Used if this turns out to 
                                      // be the innermost call in a BUF rooted nest (LEAF)
  
  // When entering deal for the first time, nbgive=nbtake=0
  FrStack[Top]->nbgive = 0;
  FrStack[Top]->nbtake = 0;

  // If the Deal is in BUF mode, there's no nesting to get the data
  if (dmode==BUF) { 
    if (DEBUGDV) printf("BUF mode\n");
    FrStack[Top]->nestlevel = 1;
  }
  // Otherwise increase the nesting level by 1
  else {
    if (DEBUGDV) printf("STRM mode\n");
    FrStack[Top]->nestlevel = FrStack[Top-1]->nestlevel + 1;
  }

  ROOT = Top - FrStack[Top]->nestlevel + 1;  

  // Copy the parameters into the FrStack
  FrStack[Top]->in          = in;
  FrStack[Top]->inmul       = inmul;
  FrStack[Top]->inlens      = &inlen; 
  FrStack[Top]->inty        = inty;
  FrStack[Top]->inbufspr    = inspr;
  FrStack[Top]->inbufcount  = 0;
  FrStack[Top]->out         = out;
  outlenptr = &outlen;
  FrStack[Top]->outlenP     = &outlenptr;
  FrStack[Top]->outmulP     = &outmul;
  FrStack[Top]->outty       = outty;
  FrStack[Top]->outbuffsz   = outbuffsz;
  FrStack[Top]->outbufspr   = outspr;
  FrStack[Top]->outbufcount = 0;

  FrStack[Top]->parentcomm  = comm;
  MPI_Comm_size(comm, &FrStack[Top]->parentcommsize);
  MPI_Comm_rank(comm, &FrStack[Top]->parentcommrank);
 
  // Copy informations relative to deal
  MPI_Comm_dup (comm, &FrStack[Top]->context.d.tofarmercomm);
  MPI_Comm_dup (comm, &FrStack[Top]->context.d.fromfarmercomm);
  FrStack[Top]->context.d.nw     = nw;
  FrStack[Top]->context.d.inspr  = inspr;
  FrStack[Top]->context.d.outspr = outspr;

  // Size of the first item
  MPI_Type_size(inty, &FrStack[Top]->initemsz);
  MPI_Type_size(outty, &FrStack[Top]->outitemsz);

  // Creating the new communicators of subgroups relating to "col"
  MPI_Comm_split(comm, col, 0, &FrStack[Top]->mycomm);
  MPI_Comm_rank(FrStack[Top]->mycomm, &FrStack[Top]->myrank);
  MPI_Comm_size(FrStack[Top]->mycomm, &FrStack[Top]->mycommsize);

  // Duplicating the parent communicator for other purpose
  MPI_Comm_dup(FrStack[Top]->parentcomm, &(FrStack[Top]->allcomm1));
  MPI_Comm_dup(FrStack[Top]->parentcomm, &(FrStack[Top]->allcomm2));

  // Need to create a comm just with the stage root
  MPI_Comm_group(FrStack[Top]->mycomm, &tempgroup);
  MPI_Group_incl(tempgroup,1,&zero, &stagerootonlygroup);
  MPI_Comm_create(FrStack[Top]->mycomm, stagerootonlygroup, &(FrStack[Top]->stagerootonlycomm));

  // Specify the ranks in the parent communicator (from 0 to parentcommsize)
  FrStack[Top]->allranks = (int *) malloc (sizeof(int)* FrStack[Top]->parentcommsize);
  for (i=0; i<FrStack[Top]->parentcommsize; i++)
    FrStack[Top]->allranks[i] = i;

   
  // work out who's who (MPI_Allgather store the results in allcols)
  FrStack[Top]->context.d.allcols = (int *) malloc (sizeof(int)*FrStack[Top]->parentcommsize);
  MPI_Allgather((void *) &FrStack[Top]->mycol, 1, MPI_INT,
		(void *) FrStack[Top]->context.d.allcols, 1, MPI_INT, comm);

  
  // Run the deal
  
  //-------------------- Implicit interaction mode -----------------------//
  // We have to perform all the Give/Take necessary for every input
  
  if (imode == IMPL) {
    if (DEBUGDV) printf("Running deal %d in IMPL mode \n", FrStack[0]->myrank);
    
    FrStack[Top]->context.d.shutdown = FALSE;
    tempitemin = (eSkel_molecule_t *) malloc (sizeof (eSkel_molecule_t));
    tempitemout = (eSkel_molecule_t *) malloc (sizeof (eSkel_molecule_t));
    len = (int **) malloc (sizeof (int *));
     
    // Take data, call the worker function and Give the output data if there is some
    while (tempitemin->data = Take(len))  {
      tempitemin->len = *len;
      tempitemout = worker(tempitemin);
      if (tempitemout) Give (tempitemout->data, tempitemout->len);
      else Give (NULL, 0);
    }
    
    // Give the shutdown tag if i'm not the last stage
    // (this means if the first worker is not OutBuffGroup)
    //if (GetMyParentGroup()->acts[0] != OutBuffGroup(FrStack[Top]->mycalltree,ROOT)) {
    givelen = (int *) malloc (sizeof(int));
    givelen[0] = SHUTDOWNLEN;
    Give(NULL, givelen);
      //}
    
  }

  //-------------------- Explicit interaction mode -----------------------//
  // Not supported by the Deal skeleton
  else if (imode == EXPL) {
    printf("*** EXPL mode is not supported by the Deal skeleton ***\n");
    exit(0);
  }

  //-------------------- Devolved interaction mode -----------------------//
  // The activity is delegated to a persistently nested skeleton.
  // We just call the stage once, which will start this other skeleton
  else {
    if (DEBUGDV) printf("Running worker %d in DEV mode \n", FrStack[0]->myrank);
    worker(NULL);
  }


  Top--; // and give back the space to be tidy
  return;
}

